"""
Tests for the code from docs/source/example.
"""
import os
import sys

from twisted.test import proto_helpers
from twisted.trial import unittest

from ldaptor import inmemory
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols import pureldap

# We inject the examples so that we can import them.
# There is no cleanup, so this is leaving side effects.
sys.path.append(os.path.abspath("docs/source/examples"))
import ldaptor_with_upn_bind


class LDAPServerWithUPNBind(unittest.TestCase):
    """
    Tests for docs/source/examples/ldaptor_with_upn_bind.py
    """

    def setUp(self):
        self.root = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="dc=example,dc=com", attributes={"dc": "example"}
        )
        self.user = self.root.addChild(
            rdn=b"cn=bob",
            attributes={
                "objectClass": ["a", "b"],
                # Hash is for "secret".
                "userPassword": [b"{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="],
                "userPrincipalName": ["bob@ad.example.com"],
            },
        )

        server = ldaptor_with_upn_bind.LDAPServerWithUPNBind()
        server.factory = self.root
        server.transport = proto_helpers.StringTransport()
        server.connectionMade()
        self.server = server

    def checkSuccessfulBIND(self, bind_dn, password):
        """
        Do a BIND request and check that is succeeds.
        """
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn=bind_dn, auth=password), id=4
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0, matchedDN="cn=bob,dc=example,dc=com"
                ),
                id=4,
            ).toWire(),
        )

    def test_bindSuccessUPN(self):
        """
        It can authenticate based on the UPN.
        """
        self.checkSuccessfulBIND("bob@ad.example.com", b"secret")

    def test_bindSuccessDN(self):
        """
        It can still authenticate based on the normal DN.
        """
        self.checkSuccessfulBIND("cn=bob,dc=example,dc=com", b"secret")

    def test_bindBadPassword(self):
        """
        When password don't match the BIND fails.
        """
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn="bob@ad.example.com", auth="invalid"),
                id=734,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=734,
            ).toWire(),
        )



"""Find an available uidNumber/gidNumber/other similar number."""

from ldaptor.protocols import pureldap


class freeNumberGuesser:
    def __init__(self, makeAGuess, min=None, max=None):
        self.makeAGuess = makeAGuess
        self.min = min
        if self.min is None:
            self.min = 0
        self.max = max

    def startGuessing(self):
        d = self.makeAGuess(self.min)
        d.addCallback(self._nextGuess, self.min)
        return d

    def _nextGuess(self, found, lastGuess):
        if found:
            self.min = lastGuess
        else:
            self.max = lastGuess

        if self.max == self.min or self.max == self.min + 1:
            return self.max

        max = self.max
        if max is None:
            max = self.min + 1000

        guess = (max + self.min) // 2
        d = self.makeAGuess(guess)
        d.addCallback(self._nextGuess, guess)
        return d


class ldapGuesser:
    def __init__(self, ldapObject, numberType):
        self.numberType = numberType
        self.ldapObject = ldapObject

    def guess(self, num):
        d = self.ldapObject.search(
            filterObject=pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value=self.numberType),
                assertionValue=pureldap.LDAPAssertionValue(value=str(num)),
            ),
            sizeLimit=1,
        )
        d.addCallback(lambda results: len(results))
        return d


def getFreeNumber(ldapObject, numberType, min=None, max=None):
    g = freeNumberGuesser(ldapGuesser(ldapObject, numberType).guess, min=min, max=max)
    return g.startGuessing()



import os
from zope.interface import Interface, implements
from twisted.internet import reactor, defer
from twisted.cred import portal, checkers
from nevow import rend, appserver, inevow, guard, loaders
from formless import annotate, webform
import nevow.flat

from ldaptor.protocols.ldap import (
    ldapclient,
    ldapsyntax,
    ldapconnector,
    distinguishedname,
)
from ldaptor import ldapfilter
from ldaptor.protocols import pureldap


class ILDAPConfig(Interface):
    """Addressbook configuration retrieval."""

    def getBaseDN(self):
        """Get the LDAP base DN, as a DistinguishedName."""

    def getServiceLocationOverrides(self):
        """
        Get the LDAP service location overrides, as a mapping of
        DistinguishedName to (host, port) tuples.
        """


class LDAPConfig:
    implements(ILDAPConfig)

    def __init__(self, baseDN, serviceLocationOverrides=None):
        self.baseDN = distinguishedname.DistinguishedName(baseDN)
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k, v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn] = v

    def getBaseDN(self):
        return self.baseDN

    def getServiceLocationOverrides(self):
        return self.serviceLocationOverrides


class LDAPSearchFilter(annotate.String):
    def coerce(self, *a, **kw):
        val = super().coerce(*a, **kw)
        try:
            f = ldapfilter.parseFilter(val)
        except ldapfilter.InvalidLDAPFilter as e:
            raise annotate.InputError(
                "{!r} is not a valid LDAP search filter: {}".format(val, e)
            )
        return f


class IAddressBookSearch(annotate.TypedInterface):
    search = LDAPSearchFilter(label="Search filter")


class CurrentSearch:
    implements(IAddressBookSearch, inevow.IContainer)
    search = None

    def child(self, context, name):
        if name == "searchFilter":
            return self.search
        if name != "results":
            return None
        config = context.locate(ILDAPConfig)

        c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d = c.connectAnonymously(
            config.getBaseDN(), config.getServiceLocationOverrides()
        )

        def _search(proto, base, searchFilter):
            baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=base)
            d = baseEntry.search(filterObject=searchFilter)
            return d

        d.addCallback(_search, config.getBaseDN(), self.search)
        return d


def LDAPFilterSerializer(original, context):
    return original.asText()


# TODO need to make this pretty some day.
for c in [
    pureldap.LDAPFilter_and,
    pureldap.LDAPFilter_or,
    pureldap.LDAPFilter_not,
    pureldap.LDAPFilter_substrings,
    pureldap.LDAPFilter_equalityMatch,
    pureldap.LDAPFilter_greaterOrEqual,
    pureldap.LDAPFilter_lessOrEqual,
    pureldap.LDAPFilter_approxMatch,
    pureldap.LDAPFilter_present,
    pureldap.LDAPFilter_extensibleMatch,
]:
    nevow.flat.registerFlattener(LDAPFilterSerializer, c)


class AddressBookResource(rend.Page):
    addSlash = True

    docFactory = loaders.xmlfile(
        "searchform.xhtml", templateDir=os.path.split(os.path.abspath(__file__))[0]
    )

    def configurable_(self, context):
        request = context.locate(inevow.IRequest)
        i = request.session.getComponent(IAddressBookSearch)
        if i is None:
            i = CurrentSearch()
            request.session.setComponent(IAddressBookSearch, i)
        return i

    def data_search(self, context, data):
        d = defer.maybeDeferred(self.locateConfigurable, context, "")

        def cb(configurable):
            return configurable.original

        d.addCallback(cb)
        return d

    def child_form_css(self, request):
        return webform.defaultCSS

    def render_input(self, context, data):
        return webform.renderForms()

    def render_haveSearch(self, context, data):
        r = context.tag.allPatterns(str(data.search is not None))
        return context.tag.clear()[r]

    def render_searchFilter(self, context, data):
        return data.asText()


class AddressBookRealm:
    implements(portal.IRealm)

    def __init__(self, resource):
        self.resource = resource

    def requestAvatar(self, avatarId, mind, *interfaces):
        if inevow.IResource not in interfaces:
            raise NotImplementedError("no interface")
        return (inevow.IResource, self.resource, lambda: None)


def getSite(config):
    form = AddressBookResource()
    form.remember(config, ILDAPConfig)
    realm = AddressBookRealm(form)
    site = appserver.NevowSite(
        guard.SessionWrapper(portal.Portal(realm, [checkers.AllowAnonymousAccess()]))
    )
    return site



from twisted.internet import defer
from ldaptor import numberalloc
from ldaptor.protocols.ldap import ldapsyntax, autofill


class Autofill_posix:  # TODO baseclass
    def __init__(self, baseDN, freeNumberGetter=numberalloc.getFreeNumber):
        self.baseDN = baseDN
        self.freeNumberGetter = freeNumberGetter

    def _cb_gotNumbers(self, r, ldapObject):
        uid, gid = r

        ok, val = uid
        if not ok:
            val.trap()
        ldapObject["uidNumber"] = [str(val)]

        ok, val = gid
        if not ok:
            val.trap()
        ldapObject["gidNumber"] = [str(val)]

    def start(self, ldapObject):
        assert "objectClass" in ldapObject
        if "posixAccount" not in ldapObject["objectClass"]:
            raise autofill.ObjectMissingObjectClassException(ldapObject)

        assert "loginShell" not in ldapObject
        ldapObject["loginShell"] = ["/bin/sh"]

        baseObject = ldapsyntax.LDAPEntry(client=ldapObject.client, dn=self.baseDN)
        d1 = self.freeNumberGetter(baseObject, "uidNumber", min=1000)

        d2 = self.freeNumberGetter(baseObject, "gidNumber", min=1000)

        d = defer.DeferredList([d1, d2], fireOnOneErrback=1)

        # silence the log
        d1.addErrback(lambda x: None)
        d2.addErrback(lambda x: None)

        d.addCallback(self._cb_gotNumbers, ldapObject)
        return d

    def notify(self, ldapObject, attributeType):
        pass



"""
Test cases for ldaptor.protocols.ldap.ldapfilter module.
"""

from twisted.trial import unittest
from ldaptor.protocols import pureldap
from ldaptor import ldapfilter


class RFC2254Examples(unittest.TestCase):
    def test_cn(self):
        text = "(cn=Babs Jensen)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value="Babs Jensen"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_not_cn(self):
        text = "(!(cn=Tim Howes))"
        filt = pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                assertionValue=pureldap.LDAPAssertionValue(value="Tim Howes"),
            )
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_and_or(self):
        text = "(&(objectClass=Person)(|(sn=Jensen)(cn=Babs J*)))"
        filt = pureldap.LDAPFilter_and(
            [
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription(
                        value="objectClass"
                    ),
                    assertionValue=pureldap.LDAPAssertionValue(value="Person"),
                ),
                pureldap.LDAPFilter_or(
                    [
                        pureldap.LDAPFilter_equalityMatch(
                            attributeDesc=pureldap.LDAPAttributeDescription(value="sn"),
                            assertionValue=pureldap.LDAPAssertionValue(value="Jensen"),
                        ),
                        pureldap.LDAPFilter_substrings(
                            type="cn",
                            substrings=[
                                pureldap.LDAPFilter_substrings_initial(value="Babs J")
                            ],
                        ),
                    ]
                ),
            ]
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_substrings(self):
        text = "(o=univ*of*mich*)"
        filt = pureldap.LDAPFilter_substrings(
            type="o",
            substrings=[
                pureldap.LDAPFilter_substrings_initial(value="univ"),
                pureldap.LDAPFilter_substrings_any(value="of"),
                pureldap.LDAPFilter_substrings_any(value="mich"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_1(self):
        text = "(cn:1.2.3.4.5:=Fred Flintstone)"
        filt = pureldap.LDAPFilter_extensibleMatch(
            type="cn",
            dnAttributes=False,
            matchingRule="1.2.3.4.5",
            matchValue="Fred Flintstone",
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_2(self):
        text = "(sn:dn:2.4.6.8.10:=Barney Rubble)"
        filt = pureldap.LDAPFilter_extensibleMatch(
            type="sn",
            dnAttributes=True,
            matchingRule="2.4.6.8.10",
            matchValue="Barney Rubble",
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_3(self):
        text = "(o:dn:=Ace Industry)"
        filt = pureldap.LDAPFilter_extensibleMatch(
            type="o",
            dnAttributes=True,
            matchingRule=None,
            matchValue="Ace Industry",
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_4(self):
        text = "(:dn:2.4.6.8.10:=Dino)"
        filt = pureldap.LDAPFilter_extensibleMatch(
            type=None,
            dnAttributes=True,
            matchingRule="2.4.6.8.10",
            matchValue="Dino",
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_5(self):
        text = "(cn:1.2.3.4.5:=Fred Flintstone)"
        filt = pureldap.LDAPFilter_extensibleMatch(
            type="cn",
            dnAttributes=None,
            matchingRule="1.2.3.4.5",
            matchValue="Fred Flintstone",
        )
        self.assertEqual(filt.asText(), text)

    def test_escape_parens(self):
        text = r"(o=Parens R Us \28for all your parenthetical needs\29)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="o"),
            assertionValue=pureldap.LDAPAssertionValue(
                value="Parens R Us (for all your parenthetical needs)"
            ),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_escape_asterisk(self):
        text = r"(cn=*\2A*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any(value="*"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text.lower())

    def test_escape_backslash(self):
        text = r"(filename=C:\5cMyFile)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="filename"),
            assertionValue=pureldap.LDAPAssertionValue(value=r"C:\MyFile"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_escape_binary(self):
        text = r"(bin=\00\00\00\04)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="bin"),
            assertionValue=pureldap.LDAPAssertionValue(value="\00\00\00\04"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)

    def test_escape_utf8(self):
        text = r"(sn=Lu\c4\8di\c4\87)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="sn"),
            assertionValue=pureldap.LDAPAssertionValue(value="Lu\xc4\x8di\xc4\x87"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        # self.assertEqual(filt.asText(), text)


class TestValid(unittest.TestCase):
    def test_item_present(self):
        text = r"(cn=*)"
        filt = pureldap.LDAPFilter_present(value="cn")
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_simple(self):
        text = r"(cn=foo)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value="foo"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_init(self):
        text = r"(cn=foo*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_final(self):
        text = r"(cn=*foo)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_final("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_any(self):
        text = r"(cn=*foo*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_aa(self):
        text = r"(cn=*foo*bar*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_ia(self):
        text = r"(cn=foo*bar*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaa(self):
        text = r"(cn=foo*bar*baz*)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_any("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_if(self):
        text = r"(cn=foo*bar)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_final("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaf(self):
        text = r"(cn=foo*bar*baz)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_final("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaaf(self):
        text = r"(cn=foo*bar*baz*quux)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_any("baz"),
                pureldap.LDAPFilter_substrings_final("quux"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_af(self):
        text = r"(cn=*foo*bar)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_final("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_aaf(self):
        text = r"(cn=*foo*bar*baz)"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_final("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_not_item(self):
        text = r"(!(cn=foo))"
        filt = pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                assertionValue=pureldap.LDAPAssertionValue(value="foo"),
            )
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_or_item(self):
        text = r"(|(cn=foo)(cn=bar))"
        filt = pureldap.LDAPFilter_or(
            [
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                    assertionValue=pureldap.LDAPAssertionValue(value="foo"),
                ),
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                    assertionValue=pureldap.LDAPAssertionValue(value="bar"),
                ),
            ]
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_and_item(self):
        text = r"(&(cn=foo)(cn=bar))"
        filt = pureldap.LDAPFilter_and(
            [
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                    assertionValue=pureldap.LDAPAssertionValue(value="foo"),
                ),
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
                    assertionValue=pureldap.LDAPAssertionValue(value="bar"),
                ),
            ]
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_andornot(self):
        text = r"(&(!(|(cn=foo)(cn=bar)))(sn=a*b*c*d))"
        filt = pureldap.LDAPFilter_and(
            [
                pureldap.LDAPFilter_not(
                    pureldap.LDAPFilter_or(
                        [
                            pureldap.LDAPFilter_equalityMatch(
                                attributeDesc=pureldap.LDAPAttributeDescription(
                                    value="cn"
                                ),
                                assertionValue=pureldap.LDAPAssertionValue(value="foo"),
                            ),
                            pureldap.LDAPFilter_equalityMatch(
                                attributeDesc=pureldap.LDAPAttributeDescription(
                                    value="cn"
                                ),
                                assertionValue=pureldap.LDAPAssertionValue(value="bar"),
                            ),
                        ]
                    )
                ),
                pureldap.LDAPFilter_substrings(
                    type="sn",
                    substrings=[
                        pureldap.LDAPFilter_substrings_initial("a"),
                        pureldap.LDAPFilter_substrings_any("b"),
                        pureldap.LDAPFilter_substrings_any("c"),
                        pureldap.LDAPFilter_substrings_final("d"),
                    ],
                ),
            ]
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_whitespace_beforeCloseParen(self):
        text = r"(cn=foo )"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value="foo "),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_whitespace_afterEq(self):
        text = r"(cn= foo)"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value=" foo"),
        )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)


class TestInvalid(unittest.TestCase):
    def test_closeParen_1(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter,
            ldapfilter.parseFilter,
            "(&(|(mail=)@*)(uid=)))(mail=*))",
        )

    def test_closeParen_2(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(|(mail=)@*)(uid=)))"
        )

    def test_closeParen_3(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(mail=)@*)"
        )

    def test_closeParen_4(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(uid=))"
        )

    def test_openParen_1(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter,
            ldapfilter.parseFilter,
            "(&(|(mail=(@*)(uid=())(mail=*))",
        )

    def test_openParen_2(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(|(mail=(@*)(uid=())"
        )

    def test_openParen_3(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(mail=(@*)"
        )

    def test_openParen_4(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, "(uid=()"
        )

    def test_whitespace_leading(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, r" (cn=foo)"
        )

    def test_whitespace_trailing(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, r"(cn=foo) "
        )

    def test_whitespace_afterOpenParen(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, r"( cn=foo)"
        )

    def test_whitespace_beforeEq(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, r"(cn =foo)"
        )


class TestMaybeSubstring(unittest.TestCase):
    def test_item_present(self):
        text = r"*"
        filt = pureldap.LDAPFilter_present(value="cn")
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_simple(self):
        text = r"foo"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value="foo"),
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_init(self):
        text = r"foo*"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_final(self):
        text = r"*foo"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_final("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_any(self):
        text = r"*foo*"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_aa(self):
        text = r"*foo*bar*"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_ia(self):
        text = r"foo*bar*"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_iaa(self):
        text = r"foo*bar*baz*"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_any("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_if(self):
        text = r"foo*bar"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_final("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_iaf(self):
        text = r"foo*bar*baz"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_final("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_iaaf(self):
        text = r"foo*bar*baz*quux"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_initial("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_any("baz"),
                pureldap.LDAPFilter_substrings_final("quux"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_af(self):
        text = r"*foo*bar"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_final("bar"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_item_substring_aaf(self):
        text = r"*foo*bar*baz"
        filt = pureldap.LDAPFilter_substrings(
            type="cn",
            substrings=[
                pureldap.LDAPFilter_substrings_any("foo"),
                pureldap.LDAPFilter_substrings_any("bar"),
                pureldap.LDAPFilter_substrings_final("baz"),
            ],
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)

    def test_escape_simple(self):
        text = r"f\2aoo(bar"
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value="cn"),
            assertionValue=pureldap.LDAPAssertionValue(value="f*oo(bar"),
        )
        self.assertEqual(ldapfilter.parseMaybeSubstring("cn", text), filt)


class TestWhitespace(unittest.TestCase):
    def test_escape(self):
        self.assertRaises(
            ldapfilter.InvalidLDAPFilter, ldapfilter.parseFilter, r"(cn=\ 61)"
        )



"""
Test cases for the ldaptor.samba.smbpassword module.
"""

from twisted.trial import unittest
from ldaptor.samba import smbpassword
from ldaptor import config


class TestNTHash(unittest.TestCase):
    """
    Unit tests for smbpassword.nthash.
    """

    knownValues = [  # password, expected_result
        (b"", b"31D6CFE0D16AE931B73C59D7E0C089C0"),
        (b"foo", b"AC8E657F83DF82BEEA5D43BDAF7800CC"),
        (127 * b"x", b"25900FAB94A048BCF438615217776562"),
        (128 * b"x", b"65681023D0CB5E7E96FF662150EF060D"),
        (129 * b"x", b"65681023D0CB5E7E96FF662150EF060D"),
        (1000 * b"x", b"65681023D0CB5E7E96FF662150EF060D"),
    ]

    def testKnownValues(self):
        """nthash(...) gives known results"""
        self.assertEqual(
            self.knownValues, [(p, smbpassword.nthash(p)) for p, _ in self.knownValues]
        )


class TestLMHash(unittest.TestCase):
    """
    Unit tests for smbpassword.lmhash.
    """

    knownValues = [  # password, expected_result
        (b"", b"AAD3B435B51404EEAAD3B435B51404EE"),
        (b"foo", b"5BFAFBEBFB6A0942AAD3B435B51404EE"),
        (13 * b"x", b"3AA62DBBEFDB676366B4159AF5A7C45C"),
        (14 * b"x", b"3AA62DBBEFDB67633AA62DBBEFDB6763"),
        (15 * b"x", b"3AA62DBBEFDB67633AA62DBBEFDB6763"),
        (100 * b"x", b"3AA62DBBEFDB67633AA62DBBEFDB6763"),
        (b"1234567abcdefg", b"0182BD0BD4444BF8E0C510199CC66ABD"),
        (b"XXXXXXXabcdefg", b"3AA62DBBEFDB6763E0C510199CC66ABD"),
        (b"1234567XXXXXXX", b"0182BD0BD4444BF83AA62DBBEFDB6763"),
    ]

    def testKnownValues(self):
        """
        When lanman password are enabled in the configuration, they will
        return
        """
        cfg = config.loadConfig()
        cfg.set("samba", "use-lmhash", "yes")
        self.assertEqual(
            self.knownValues, [(p, smbpassword.lmhash(p)) for p, _ in self.knownValues]
        )

    def testLMHashConfiguationDisabled(self):
        """
        When the lanman passwords are disabled from the configuration, it will
        return the disabled password.
        """
        cfg = config.loadConfig()
        cfg.set("samba", "use-lmhash", "no")

        result = smbpassword.lmhash(b"any-pass")

        self.assertEqual(32 * b"X", result)



import base64

from twisted.protocols import basic
from twisted.internet import protocol

from ldaptor import entry


class LDIFParseError(Exception):
    """Error parsing LDIF"""

    def __str__(self):
        s = self.__doc__
        if self.args:
            s = ": ".join([s] + [str(x) for x in self.args])
        return s + "."


class LDIFLineWithoutSemicolonError(LDIFParseError):
    """LDIF line without semicolon seen"""


class LDIFEntryStartsWithNonDNError(LDIFParseError):
    """LDIF entry starts with a non-DN line"""


class LDIFEntryStartsWithSpaceError(LDIFParseError):
    """Invalid LDIF value format"""


class LDIFVersionNotANumberError(LDIFParseError):
    """Non-numeric LDIF version number"""


class LDIFUnsupportedVersionError(LDIFParseError):
    """LDIF version not supported"""


class LDIFTruncatedError(LDIFParseError):
    """LDIF appears to be truncated"""


HEADER = b"HEADER"
WAIT_FOR_DN = b"WAIT_FOR_DN"
IN_ENTRY = b"IN_ENTRY"


class LDIF(basic.LineReceiver):
    delimiter = b"\n"
    mode = HEADER

    dn = None
    data = None
    lastLine = None

    version = None

    def logicalLineReceived(self, line):
        if line.startswith(b"#"):
            # comments are allowed everywhere
            return
        getattr(self, "state_" + self.mode.decode("ascii"))(line)

    def lineReceived(self, line):
        if line.startswith(b" "):
            if self.lastLine is None:
                raise LDIFEntryStartsWithSpaceError()
            self.lastLine = self.lastLine + line[1:]
        else:
            if self.lastLine is not None:
                self.logicalLineReceived(self.lastLine)
            self.lastLine = line
            if line == b"":
                self.logicalLineReceived(line)
                self.lastLine = None

    def parseValue(self, val):
        if val.startswith(b":"):
            return base64.decodestring(val[1:].lstrip(b" "))
        elif val.startswith(b"<"):
            raise NotImplementedError()
        else:
            return val.lstrip(b" ")

    def _parseLine(self, line):
        try:
            key, val = line.split(b":", 1)
        except ValueError:
            # unpack list of wrong size
            # -> invalid input data
            raise LDIFLineWithoutSemicolonError(line)
        val = self.parseValue(val)
        return key, val

    def state_HEADER(self, line):
        key, val = self._parseLine(line)
        self.mode = WAIT_FOR_DN

        if key != b"version":
            self.logicalLineReceived(line)
        else:
            try:
                version = int(val)
            except ValueError:
                raise LDIFVersionNotANumberError(val)
            self.version = version
            if version > 1:
                raise LDIFUnsupportedVersionError(version)

    def state_WAIT_FOR_DN(self, line):
        assert self.dn is None, "self.dn must not be set when waiting for DN"
        assert self.data is None, "self.data must not be set when waiting for DN"
        if line == b"":
            # too many empty lines, but be tolerant
            return

        key, val = self._parseLine(line)

        if key.upper() != b"DN":
            raise LDIFEntryStartsWithNonDNError(line)

        self.dn = val
        self.data = {}
        self.mode = IN_ENTRY

    def state_IN_ENTRY(self, line):
        assert self.dn is not None, "self.dn must be set when in entry"
        assert self.data is not None, "self.data must be set when in entry"

        if line == b"":
            # end of entry
            self.mode = WAIT_FOR_DN
            o = entry.BaseLDAPEntry(dn=self.dn, attributes=self.data)
            self.dn = None
            self.data = None
            self.gotEntry(o)
            return

        key, val = self._parseLine(line)

        if not key in self.data:
            self.data[key] = []

        self.data[key].append(val)

    def gotEntry(self, obj):
        pass

    def connectionLost(self, reason=protocol.connectionDone):
        if self.mode != WAIT_FOR_DN:
            raise LDIFTruncatedError(reason)



"""
    Test cases for ldaptor.protocols.ldap.ldif module
"""

import base64

from twisted.trial import unittest

from ldaptor.protocols.ldap.ldif import attributeAsLDIF, asLDIF, manyAsLDIF


def encode(value):
    return b"".join(base64.encodestring(value).split(b"\n"))


class WireableObject:
    """
    Object with bytes representation as a constant toWire value
    """

    def toWire(self):
        return b"wire"


class AttributeAsLDIFTests(unittest.TestCase):
    """
    Converting pairs of attribute keys and values to LDIF.
    The result is a byte string with key and value
    separated by a semicolon and a space.
    In several special cases value is base64 encoded and separated
    from key by two semicolons and a space.
    """

    def test_byte_string(self):
        """Key and value are byte strings"""
        result = attributeAsLDIF(b"some key", b"some value")
        self.assertEqual(result, b"some key: some value\n")

    def test_unicode_string(self):
        """Key and value are unicode strings"""
        result = attributeAsLDIF("another key", "another value")
        self.assertEqual(result, b"another key: another value\n")

    def test_wireable_object(self):
        """Value is an object with toWire method returning its bytes representation"""
        result = attributeAsLDIF("dn", WireableObject())
        self.assertEqual(result, b"dn: wire\n")

    def test_startswith_special_character(self):
        """
        Value is a string starting with one of the reserved characters.
        Returned value is base64 encoded.
        """
        for c in b"\0", b"\n", b"\r", b" ", b":", b"<":

            value = c + b"value"
            result = attributeAsLDIF(b"key", value)
            self.assertEqual(result, b"key:: %s\n" % encode(value))

    def test_endswith_special_character(self):
        """
        Value is a string ending with one of the reserved characters.
        Returned value is base64 encoded.
        """
        for c in b"\0", b"\n", b"\r", b" ":

            value = b"value" + c
            result = attributeAsLDIF(b"key", value)
            self.assertEqual(result, b"key:: %s\n" % encode(value))

    def test_contains_special_characters(self):
        """
        Value is a string with one of the reserved characters
        somewhere in its middle.
        Returned value is base64 encoded.
        """
        for c in b"\0", b"\n", b"\r":

            value = b"foo" + c + b"bar"
            result = attributeAsLDIF(b"key", value)
            self.assertEqual(result, b"key:: %s\n" % encode(value))

    def test_contains_nonprintable_characters(self):
        """
        Value is a string containing nonprintable characters.
        Returned value is base64 encoded.
        """
        result = attributeAsLDIF(b"key", b"val\xFFue")
        self.assertEqual(result, b"key:: %s\n" % encode(b"val\xFFue"))


class AsLDIFTests(unittest.TestCase):
    """
    Converting LDAP objects to LDIF.
    Object consists of DN and pairs of attribute keys and lists of values.
    The result is a number of lines in LDIF format for every key/value pair
    including DN.
    """

    def test_byte_string(self):
        """DN and attribute keys and values are byte strings"""
        attributes = [
            (b"key1", [b"value11", b"value12"]),
            (b"key2", [b"value21", b"value22"]),
        ]
        result = asLDIF(b"entry", attributes)
        self.assertEqual(
            result,
            b"""\
dn: entry
key1: value11
key1: value12
key2: value21
key2: value22

""",
        )

    def test_unicode_string(self):
        """DN and attribute keys and values are unicode string"""
        attributes = [
            ("key1", ["value11", "value12"]),
            ("key2", ["value21", "value22"]),
        ]
        result = asLDIF("entry", attributes)
        self.assertEqual(
            result,
            b"""\
dn: entry
key1: value11
key1: value12
key2: value21
key2: value22

""",
        )

    def test_wireable_object(self):
        """
        DN and attribute values are objects with
        toWire method returning bytes representation
        """
        attributes = [
            (b"key", [WireableObject()]),
        ]
        result = asLDIF(WireableObject(), attributes)
        self.assertEqual(result, b"dn: wire\nkey: wire\n\n")


class ManyAsLDIFTests(unittest.TestCase):
    """
    Converting multiple LDAP objects to LDIF.
    Every object consists of pair of DN and pairs of attribute keys and lists of values.
    The result is a number of blocks representing each object in LDIF format
    separated by an empty lines.
    The result contains a header with version number separated from other blocks
    by an empty line.
    """

    def test_multiple_objects(self):
        objects = [
            (
                b"object1",
                (
                    (b"foo1", (b"value11", b"value12")),
                    (b"foo2", (b"value21", b"value22")),
                ),
            ),
            (
                b"object2",
                (
                    (b"bar1", (b"value31", b"value32")),
                    (b"bar2", (b"value41", b"value42")),
                ),
            ),
        ]
        result = manyAsLDIF(objects)
        self.assertEqual(
            result,
            b"""\
version: 1

dn: object1
foo1: value11
foo1: value12
foo2: value21
foo2: value22

dn: object2
bar1: value31
bar1: value32
bar2: value41
bar2: value42

""",
        )



"""LDAP protocol proxy server"""

from twisted.internet import reactor, defer
from ldaptor.protocols.ldap import ldapserver, ldapconnector, ldapclient
from ldaptor.protocols import pureldap


class Proxy(ldapserver.BaseLDAPServer):
    protocol = ldapclient.LDAPClient

    client = None
    waitingConnect = []
    unbound = False

    def __init__(self, config):
        """
        Initialize the object.

        @param config: The configuration.
        @type config: ldaptor.interfaces.ILDAPConfig
        """
        ldapserver.BaseLDAPServer.__init__(self)
        self.config = config

    def _whenConnected(self, fn, *a, **kw):
        if self.client is None:
            d = defer.Deferred()
            self.waitingConnect.append((d, fn, a, kw))
            return d
        else:
            return defer.maybeDeferred(fn, *a, **kw)

    def _cbConnectionMade(self, proto):
        self.client = proto
        while self.waitingConnect:
            d, fn, a, kw = self.waitingConnect.pop(0)
            d2 = defer.maybeDeferred(fn, *a, **kw)
            d2.chainDeferred(d)

    def _clientQueue(self, request, controls, reply):
        # TODO controls
        if request.needs_answer:
            self.client.send_multiResponse(request, self._gotResponse, reply)
            # TODO handle errbacks from the deferred above
        else:
            self.client.send_noResponse(request)

    def _gotResponse(self, response, reply):
        reply(response)

        # TODO this is ugly
        return isinstance(
            response,
            (
                pureldap.LDAPSearchResultDone,
                pureldap.LDAPBindResponse,
            ),
        )

    def _failConnection(self, reason):
        # TODO self.loseConnection()
        return reason  # TODO

    def connectionMade(self):
        clientCreator = ldapconnector.LDAPClientCreator(reactor, self.protocol)
        d = clientCreator.connect(
            dn="", overrides=self.config.getServiceLocationOverrides()
        )
        d.addCallback(self._cbConnectionMade)
        d.addErrback(self._failConnection)

        ldapserver.BaseLDAPServer.connectionMade(self)

    def connectionLost(self, reason):
        assert self.client is not None
        if self.client.connected:
            if not self.unbound:
                self.client.unbind()
                self.unbound = True
            else:
                self.client.transport.loseConnection()
        self.client = None
        ldapserver.BaseLDAPServer.connectionLost(self, reason)

    def _handleUnknown(self, request, controls, reply):
        self._whenConnected(self._clientQueue, request, controls, reply)
        return None

    def handleUnknown(self, request, controls, reply):
        d = defer.succeed(request)
        d.addCallback(self._handleUnknown, controls, reply)
        return d

    def handle_LDAPUnbindRequest(self, request, controls, reply):
        self.unbound = True
        self.handleUnknown(request, controls, reply)


if __name__ == "__main__":
    """
    Demonstration LDAP proxy; passes all requests to localhost:389.
    """
    from twisted.internet import protocol
    from twisted.python import log
    import sys

    log.startLogging(sys.stderr)

    factory = protocol.ServerFactory()
    factory.protocol = lambda: Proxy(
        overrides={
            "": ("localhost", 389),
        }
    )
    reactor.listenTCP(10389, factory)
    reactor.run()



#!/usr/bin/env python3

import codecs
import os
import re

from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as f:
        return f.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


if __name__ == "__main__":
    setup(version=find_version("ldaptor", "__init__.py"))



#! /usr/bin/env python

import argparse
import sys

from twisted.internet import defer
from twisted.internet.endpoints import clientFromString, connectProtocol
from twisted.internet.task import react
from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry
from ldaptor.protocols import pureber


@defer.inlineCallbacks
def onConnect(client, args):
    binddn = args.bind_dn
    bindpw = args.passwd_file.read().strip()
    if args.start_tls:
        yield client.startTLS()
    try:
        yield client.bind(binddn, bindpw)
    except Exception as ex:
        print(ex)
        raise
    page_size = args.page_size
    cookie = ""
    page = 1
    count = 0
    while True:
        results, cookie = yield process_entry(
            client, args, args.filter, page_size=page_size, cookie=cookie
        )
        count += len(results)
        print("Page {}".format(page))
        display_results(results)
        if len(cookie) == 0:
            break
        page += 1
    print("There were {} results returned in total.".format(count))


@defer.inlineCallbacks
def process_entry(client, args, search_filter, page_size=100, cookie=""):
    basedn = args.base_dn
    control_value = pureber.BERSequence(
        [
            pureber.BERInteger(page_size),
            pureber.BEROctetString(cookie),
        ]
    )
    controls = [("1.2.840.113556.1.4.319", None, control_value)]
    o = LDAPEntry(client, basedn)
    results, resp_controls = yield o.search(
        filterText=search_filter,
        attributes=["dn"],
        controls=controls,
        return_controls=True,
    )
    cookie = get_paged_search_cookie(resp_controls)
    defer.returnValue((results, cookie))


def display_results(results):
    for entry in results:
        print(entry.dn.getText())


def get_paged_search_cookie(controls):
    """
    Input: semi-parsed controls list from LDAP response;
    list of tuples (controlType, criticality, controlValue).
    Parses the controlValue and returns the cookie as a byte string.
    """
    control_value = controls[0][2]
    ber_context = pureber.BERDecoderContext()
    ber_seq, bytes_used = pureber.berDecodeObject(ber_context, control_value)
    raw_cookie = ber_seq[1]
    cookie = raw_cookie.value
    return cookie


def onError(err):
    err.printDetailedTraceback(file=sys.stderr)


def main(reactor, args):
    endpoint_str = args.endpoint
    e = clientFromString(reactor, endpoint_str)
    d = connectProtocol(e, LDAPClient())
    d.addCallback(onConnect, args)
    d.addErrback(onError)
    return d


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="AD LDAP demo.")
    parser.add_argument(
        "endpoint",
        action="store",
        help="The Active Directory service endpoint. See "
        "https://twistedmatrix.com/documents/current/core/howto/endpoints.html#clients",
    )
    parser.add_argument(
        "bind_dn", action="store", help="The DN to BIND to the service as."
    )
    parser.add_argument(
        "passwd_file",
        action="store",
        type=argparse.FileType("r"),
        help="A file containing the password used to log into the service.",
    )
    parser.add_argument(
        "base_dn", action="store", help="The base DN to start from when searching."
    )
    parser.add_argument("-f", "--filter", action="store", help="LDAP filter")
    parser.add_argument(
        "-p",
        "--page-size",
        type=int,
        action="store",
        default=100,
        help="Page size (default 100).",
    )
    parser.add_argument(
        "--start-tls",
        action="store_true",
        help="Request StartTLS after connecting to the service.",
    )
    args = parser.parse_args()
    react(main, [args])



"""
Test cases for ldaptor.protocols.ldap.ldapsyntax module.
"""

import re

from twisted.trial import unittest
from ldaptor import config, testutil, delta
from ldaptor.protocols.ldap import ldapclient, ldapsyntax, ldaperrors
from ldaptor.protocols import pureldap, pureber
from twisted.internet import defer
from twisted.internet import error
from twisted.python import failure
from ldaptor.testutil import LDAPClientTestDriver


class LDAPEntryTests(unittest.TestCase):
    """
    Unit tests for LDAPEntry.
    """

    def testCreation(self):
        """Creating an LDAP object should succeed."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        self.failUnlessEqual(o.dn.getText(), "cn=foo,dc=example,dc=com")
        self.failUnlessEqual(o["objectClass"], ["a", "b"])
        self.failUnlessEqual(o["aValue"], ["a"])
        self.failUnlessEqual(o["bValue"], ["b"])
        client.assertNothingSent()

    def testKeys(self):
        """Iterating over the keys of an LDAP object gives expected results."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        seen = {}
        for k in o.keys():
            assert k not in seen
            seen[k] = 1
        assert seen == {
            "objectClass": 1,
            "aValue": 1,
            "bValue": 1,
        }

    def testItems(self):
        """Iterating over the items of an LDAP object gives expected results."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        seen = {}
        for k, vs in o.items():
            assert k not in seen
            seen[k] = vs
        assert seen == {
            "objectClass": ["a", "b"],
            "aValue": ["a"],
            "bValue": ["b"],
        }

    def testIn(self):
        """Key in object gives expected results."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        assert "objectClass" in o
        assert "aValue" in o
        assert "bValue" in o
        assert "foo" not in o
        assert "" not in o
        assert None not in o

        assert "a" in o["objectClass"]
        assert "b" in o["objectClass"]
        assert "foo" not in o["objectClass"]
        assert "" not in o["objectClass"]
        assert None not in o["objectClass"]

        assert "a" in o["aValue"]
        assert "foo" not in o["aValue"]
        assert "" not in o["aValue"]
        assert None not in o["aValue"]

    def testInequalityOtherObject(self):
        """
        It is not equal with non LDAPEntry objects.
        """
        client = LDAPClientTestDriver()
        sut = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
        )

        self.assertNotEqual("dc=example,dc=com", sut)

    def testInequalityDN(self):
        """
        Entries with different DN are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=org",
        )

        self.assertNotEqual(first, second)

    def testInequalityAttributes(self):
        """
        Entries with same DN but different attributes are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={"attr_key1": ["some-value"]},
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={"attr_key2": ["some-value"]},
        )

        self.assertNotEqual(first, second)

    def testInequalityValues(self):
        """
        Entries with same DN same attributes, but different
        values for attributes are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={"attr_key1": ["some-value"]},
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={"attr_key1": ["other-value"]},
        )

        self.assertNotEqual(first, second)

    def testEquality(self):
        """
        Entries with same DN, same attributes, and same values for
        attributes equal, regardless of the order of the attributes.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "attr_key1": ["some-value"],
                "attr_key2": ["second-value"],
            },
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "attr_key2": ["second-value"],
                "attr_key1": ["some-value"],
            },
        )

        self.assertEqual(first, second)

    def testHashEqual(self):
        """
        Entries which are equal have the same hash.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
        )

        self.assertEqual(first, second)
        self.assertEqual(hash(first), hash(second))

    def testHashNotEqual(self):
        """
        Entries which are not equal have different hash values.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
        )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=org",
        )

        self.assertNotEqual(first, second)
        self.assertNotEqual(hash(first), hash(second))


class LDAPSyntaxAttributes(unittest.TestCase):
    def testAttributeSetting(self):
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        o["aValue"] = ["foo", "bar"]
        self.failUnlessEqual(o["aValue"], ["foo", "bar"])
        o["aValue"] = ["quux"]
        self.failUnlessEqual(o["aValue"], ["quux"])
        self.failUnlessEqual(o["bValue"], ["b"])
        o["cValue"] = ["thud"]
        self.failUnlessEqual(o["aValue"], ["quux"])
        self.failUnlessEqual(o["bValue"], ["b"])
        self.failUnlessEqual(o["cValue"], ["thud"])

    def testAttributeDelete(self):
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        o["aValue"] = ["quux"]
        del o["aValue"]
        del o["bValue"]
        self.failIf("aValue" in o)
        self.failIf("bValue" in o)

    def testAttributeAdd(self):
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        o["aValue"].add("foo")
        self.failUnlessEqual(o["aValue"], ["a", "foo"])

    def testAttributeItemDelete(self):
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a", "b", "c"],
                "bValue": ["b"],
            },
        )
        o["aValue"].remove("b")
        self.failUnlessEqual(o["aValue"], ["a", "c"])

    def testUndo(self):
        """Undo should forget the modifications."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
                "cValue": ["c"],
            },
        )
        o["aValue"] = ["foo", "bar"]
        o["aValue"] = ["quux"]
        del o["cValue"]
        o.undo()
        self.failUnlessEqual(o["aValue"], ["a"])
        self.failUnlessEqual(o["bValue"], ["b"])
        self.failUnlessEqual(o["cValue"], ["c"])

    def testUndoJournaling(self):
        """Journaling should still work after undo."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
                "cValue": ["c"],
            },
        )
        o["aValue"] = ["foo", "bar"]
        o["aValue"] = ["quux"]
        del o["cValue"]
        o.undo()
        o["aValue"].update(["newValue", "anotherNewValue"])
        d = o.commit()

        def cb(dummy):
            self.failUnlessEqual(o["aValue"], ["a", "newValue", "anotherNewValue"])
            self.failUnlessEqual(o["bValue"], ["b"])
            self.failUnlessEqual(o["cValue"], ["c"])
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Add("aValue", ["newValue", "anotherNewValue"]),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testUndoAfterCommit(self):
        """Undo should not undo things that have been commited."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
                "cValue": ["c"],
            },
        )
        o["aValue"] = ["foo", "bar"]
        o["bValue"] = ["quux"]
        del o["cValue"]

        d = o.commit()

        def cb(dummy):
            o.undo()
            self.failUnlessEqual(o["aValue"], ["foo", "bar"])
            self.failUnlessEqual(o["bValue"], ["quux"])
            self.failIf("cValue" in o)

        d.addCallback(cb)
        return d


class LDAPSyntaxAttributesModificationOnWire(unittest.TestCase):
    def testAdd(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
            },
        )
        o["aValue"].update(["newValue", "anotherNewValue"])

        d = o.commit()

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Add("aValue", ["newValue", "anotherNewValue"]),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testAddSeparate(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
            },
        )
        o["aValue"].add("newValue")
        o["aValue"].add("anotherNewValue")

        d = o.commit()

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Add("aValue", ["newValue"]),
                        delta.Add("aValue", ["anotherNewValue"]),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testDeleteAttribute(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
            },
        )
        o["aValue"].remove("a")

        d = o.commit()

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Delete("aValue", ["a"]),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testDeleteAllAttribute(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a1", "a2"],
                "bValue": ["b1", "b2"],
            },
        )
        del o["aValue"]
        o["bValue"].clear()

        d = o.commit()

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Delete("aValue"),
                        delta.Delete("bValue"),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testReplaceAttributes(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
            },
        )
        o["aValue"] = ["foo", "bar"]

        d = o.commit()

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace("aValue", ["foo", "bar"]),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d


class LDAPSyntaxSearch(unittest.TestCase):
    timeout = 3

    def _test_search(self, return_controls=False):
        """
        Create a test search.
        Return Deferred with no handler.
        """
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=bar,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["d", "e"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )
        d = o.search(
            filterText="(foo=a)",
            attributes=["foo", "bar"],
            return_controls=return_controls,
        )

        def cb(val):
            if return_controls:
                val = val[0]
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value="foo"),
                        assertionValue=pureldap.LDAPAssertionValue(value="a"),
                    ),
                    attributes=["foo", "bar"],
                )
            )
            self.failUnlessEqual(len(val), 2)
            self.failUnlessEqual(
                val[0],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn="cn=foo,dc=example,dc=com",
                    attributes={
                        b"foo": [b"a"],
                        b"bar": [b"b", b"c"],
                    },
                ),
            )
            self.failUnlessEqual(
                val[1],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn="cn=bar,dc=example,dc=com",
                    attributes={
                        b"foo": [b"a"],
                        b"bar": [b"d", b"e"],
                    },
                ),
            )

        d.addCallback(cb)
        return d

    def testSearch(self):
        """Test searches."""
        return self._test_search()

    def test_search_not_connected(self):
        client = ldapclient.LDAPClient()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )
        d = o.search(filterText="(foo=a)", attributes=["foo", "bar"])

        def cb_(thing):
            self.assertEqual(
                ldapclient.LDAPClientConnectionLostException, type(thing.value)
            )

        d.addErrback(cb_)
        return d

    def test_search_controls_returned(self):
        return self._test_search(return_controls=True)

    def test_search_size_limit_exceeded(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=ldaperrors.LDAPSizeLimitExceeded.resultCode,
                    matchedDN="",
                    errorMessage="Size limit exceeded.",
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )
        d = o.search(
            filterText="(foo=a)",
            attributes=["foo", "bar"],
            sizeLimit=1,
            return_controls=False,
        )

        def cb_(thing):
            self.failUnlessEqual(len(thing), 1)

        d.addCallback(cb_)
        d.addErrback(cb_)
        return d

    def testSearch_defaultAttributes(self):
        """Search without explicit list of attributes returns all attributes."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=bar,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["d", "e"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )

        d = o.search(filterText="(foo=a)")

        def cb(val):
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value="foo"),
                        assertionValue=pureldap.LDAPAssertionValue(value="a"),
                    ),
                    attributes=[],
                )
            )
            self.failUnlessEqual(len(val), 2)

            self.failUnlessEqual(
                val[0],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn="cn=foo,dc=example,dc=com",
                    attributes={
                        b"foo": [b"a"],
                        b"bar": [b"b", b"c"],
                    },
                ),
            )
            self.failUnless(val[0].complete)

            self.failUnlessEqual(
                val[1],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn="cn=bar,dc=example,dc=com",
                    attributes={
                        b"foo": [b"a"],
                        b"bar": [b"d", b"e"],
                    },
                ),
            )
            self.failUnless(val[1].complete)

        d.addCallback(cb)
        return d

    def testSearch_noAttributes(self):
        """Search with attributes=None returns no attributes."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", attributes=()
                ),
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", attributes=()
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )

        d = o.search(filterText="(foo=a)", attributes=None)

        def cb(val):
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value="foo"),
                        assertionValue=pureldap.LDAPAssertionValue(value="a"),
                    ),
                    attributes=["1.1"],
                )
            )
            self.failUnlessEqual(len(val), 2)

            self.failUnlessEqual(
                val[0],
                ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com"),
            )
            self.failIf(val[0].complete)

            self.failUnlessEqual(
                val[1],
                ldapsyntax.LDAPEntry(client=client, dn="cn=bar,dc=example,dc=com"),
            )
            self.failIf(val[1].complete)

        d.addCallback(cb)
        return d

    def testSearch_ImmediateProcessing(self):
        """Test searches with the immediate processing feature."""

        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(("bar", ["b", "c"]),),
                ),
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=bar,dc=example,dc=com",
                    attributes=(("bar", ["b", "c"]),),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
            },
        )

        seen = []

        def process(o):
            seen.append(o)

        d = o.search(filterText="(foo=a)", attributes=["bar"], callback=process)

        def cb(val):
            self.assertEqual(val, None)

            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value="foo"),
                        assertionValue=pureldap.LDAPAssertionValue(value="a"),
                    ),
                    attributes=["bar"],
                )
            )

            self.failUnlessEqual(
                seen,
                [
                    ldapsyntax.LDAPEntry(
                        client=client,
                        dn="cn=foo,dc=example,dc=com",
                        attributes={
                            b"bar": [b"b", b"c"],
                        },
                    ),
                    ldapsyntax.LDAPEntry(
                        client=client,
                        dn="cn=bar,dc=example,dc=com",
                        attributes={
                            b"bar": [b"b", b"c"],
                        },
                    ),
                ],
            )

        d.addCallback(cb)
        return d

    def testSearch_fail(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=ldaperrors.LDAPBusy.resultCode,
                    matchedDN="",
                    errorMessage="Go away",
                )
            ]
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="dc=example,dc=com")
        d = o.search(filterText="(foo=a)")

        def eb(fail):
            fail.trap(ldaperrors.LDAPBusy)
            self.assertEqual(fail.value.message, "Go away")

            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value="foo"),
                        assertionValue=pureldap.LDAPAssertionValue(value="a"),
                    ),
                )
            )

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testSearch_err(self):
        client = LDAPClientTestDriver([failure.Failure(error.ConnectionLost())])
        o = ldapsyntax.LDAPEntry(client=client, dn="dc=example,dc=com")
        d = o.search(filterText="(foo=a)")

        def eb(fail):
            fail.trap(error.ConnectionLost)

        d.addCallbacks(testutil.mustRaise, eb)
        return d


class LDAPSyntaxDNs(unittest.TestCase):
    def testDNKeyExistenceSuccess(self):
        client = LDAPClientTestDriver()
        ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "cn": ["foo"],
            },
        )


class LDAPSyntaxLDIF(unittest.TestCase):
    def testLDIFConversion(self):
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a", "b"],
                "bValue": ["c"],
            },
        )
        self.failUnlessEqual(
            o.toWire(),
            b"""dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

""",
        )


class LDAPSyntaxDelete(unittest.TestCase):
    def testDeleteInvalidates(self):
        """Deleting an LDAPEntry invalidates it."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPDelResponse(resultCode=0, matchedDN="", errorMessage=""),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a"],
            },
        )
        d = o.delete()

        def cb(dummy):
            self.failUnlessRaises(
                ldapsyntax.ObjectDeletedError, o.search, filterText="(foo=a)"
            )
            self.failUnlessRaises(ldapsyntax.ObjectDeletedError, o.get, "objectClass")

        d.addCallback(cb)
        return d

    def testDeleteOnWire(self):
        """LDAPEntry.delete should write the right data to the server."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPDelResponse(resultCode=0, matchedDN="", errorMessage=""),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a"],
            },
        )
        d = o.delete()

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPDelRequest(
                    entry="cn=foo,dc=example,dc=com",
                )
            )

        d.addCallback(cb)
        return d

    def testErrorHandling(self):
        """LDAPEntry.delete should pass LDAP errors to it's deferred."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPDelResponse(
                    resultCode=ldaperrors.LDAPBusy.resultCode,
                    matchedDN="",
                    errorMessage="Go away",
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a"],
            },
        )
        d = o.delete()

        def eb(fail):
            fail.trap(ldaperrors.LDAPBusy)
            self.assertEqual(fail.value.message, "Go away")

            client.assertSent(
                pureldap.LDAPDelRequest(
                    entry="cn=foo,dc=example,dc=com",
                )
            )

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testErrorHandling_extended(self):
        """LDAPEntry.delete should pass even non-LDAPDelResponse errors to it's deferred."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    responseName="1.3.6.1.4.1.1466.20036",
                    errorMessage="Unknown request",
                )
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a"],
            },
        )
        d = o.delete()

        def eb(fail):
            fail.trap(ldaperrors.LDAPProtocolError)
            self.assertEqual(fail.value.message, "Unknown request")

            client.assertSent(
                pureldap.LDAPDelRequest(
                    entry="cn=foo,dc=example,dc=com",
                )
            )

        d.addCallbacks(testutil.mustRaise, eb)
        return d


class LDAPSyntaxAddChild(unittest.TestCase):
    def testAddChildOnWire(self):
        """LDAPEntry.addChild should write the right data to the server."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPAddResponse(resultCode=0, matchedDN="", errorMessage=""),
            ]
        )
        sut = ldapsyntax.LDAPEntry(
            client=client,
            dn="ou=things,dc=example,dc=com",
            attributes={
                "objectClass": ["organizationalUnit"],
                "ou": ["things"],
            },
        )
        d = sut.addChild(
            rdn="givenName=Firstname+surname=Lastname",
            attributes={
                "objectClass": ["person", b"otherStuff"],
                "givenName": ["Firstname"],
                "surname": ["Lastname"],
            },
        )
        self.successResultOf(d)

        client.assertSent(
            pureldap.LDAPAddRequest(
                entry="givenName=Firstname+surname=Lastname,ou=things,dc=example,dc=com",
                attributes=[
                    (
                        pureldap.LDAPAttributeDescription("objectClass"),
                        pureber.BERSet(
                            [
                                pureldap.LDAPAttributeValue("person"),
                                pureldap.LDAPAttributeValue("otherStuff"),
                            ]
                        ),
                    ),
                    (
                        pureldap.LDAPAttributeDescription("givenName"),
                        pureber.BERSet([pureldap.LDAPAttributeValue("Firstname")]),
                    ),
                    (
                        pureldap.LDAPAttributeDescription("surname"),
                        pureber.BERSet(
                            [pureldap.LDAPAttributeValue("Lastname")],
                        ),
                    ),
                ],
            )
        )


class LDAPSyntaxContainingNamingContext(unittest.SynchronousTestCase):
    def setUp(self):
        attributes = [
            (
                "namingContexts",
                (
                    "dc=foo,dc=example",
                    "dc=example,dc=com",
                    "dc=bar,dc=example",
                ),
            )
        ]
        self.client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(objectName="", attributes=attributes),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

    def testNamingContext(self):
        """LDAPEntry.namingContext returns the naming context that contains this object (via a Deferred)."""
        o = ldapsyntax.LDAPEntry(
            client=self.client,
            dn="cn=foo,ou=bar,dc=example,dc=com",
            attributes={"objectClass": ["a"]},
        )
        d = o.namingContext()

        def cb(p):
            assert isinstance(p, ldapsyntax.LDAPEntry)
            assert p.client == o.client
            assert p.dn.getText() == "dc=example,dc=com"

            self.client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                    filter=pureldap.LDAPFilter_present("objectClass"),
                    attributes=["namingContexts"],
                )
            )

        d.addCallback(cb)
        return d

    def testNoContainingNamingContext(self):
        """LDAPEntry.namingContext raises exception if there are no naming contexts with it"""
        o = ldapsyntax.LDAPEntry(
            client=self.client,
            dn="cn=foo,dc=foo,dc=com",
            attributes={"objectClass": ["a"]},
        )
        fail = self.failureResultOf(o.namingContext())
        self.assertIsInstance(fail.value, ldapsyntax.NoContainingNamingContext)


class LDAPSyntaxPasswords(unittest.TestCase):
    def setUp(self):
        cfg = config.loadConfig()
        cfg.set("samba", "use-lmhash", "no")

    def testPasswordSetting_ExtendedOperation(self):
        """LDAPEntry.setPassword_ExtendedOperation(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword_ExtendedOperation(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
            )

        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_sambaAccount(self):
        """LDAPEntry.setPassword_Samba(newPasswd=...,
        style='sambaAccount') changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword_Samba(newPasswd=b"new", style="sambaAccount")

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "ntPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "lmPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_sambaSamAccount(self):
        """LDAPEntry.setPassword_Samba(newPasswd=..., style='sambaSamAccount') changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword_Samba(newPasswd=b"new", style="sambaSamAccount")

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "sambaNTPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "sambaLMPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_defaultStyle(self):
        """LDAPEntry.setPassword_Samba(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword_Samba(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "sambaNTPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "sambaLMPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP()
            )

        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_badStyle(self):
        """LDAPEntry.setPassword_Samba(..., style='foo') fails."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = defer.maybeDeferred(o.setPassword_Samba, newPasswd=b"new", style="foo")

        def eb(fail):
            fail.trap(RuntimeError)
            self.assertEqual(
                fail.getErrorMessage(), "Unknown samba password style 'foo'"
            )
            client.assertNothingSent()

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testPasswordSettingAll_noSamba(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_hasSamba(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo", "sambaAccount"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "ntPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "lmPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP(),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_hasSambaSam(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo", "sambaSamAccount"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "sambaNTPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "sambaLMPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP(),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_hasSamba_differentCase(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo", "saMBaAccOuNT"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "ntPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "lmPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP(),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_hasSambaSam_differentCase(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo", "sAmbASAmaccoUnt"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "sambaNTPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "sambaLMPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP(),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_maybeSamba_WillFind(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "sambaAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilterMatchAll,
                    attributes=("objectClass",),
                ),
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace(
                            "ntPassword", ["89963F5042E5041A59C249282387A622"]
                        ),
                        delta.Replace(
                            "lmPassword", ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"]
                        ),
                    ],
                ).asLDAP(),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_maybeSamba_WillNotFind(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="", attributes=[("objectClass", ("foo", "bar"))]
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword(newPasswd=b"new")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilterMatchAll,
                    attributes=("objectClass",),
                ),
            )

        d.addCallback(cb)
        return d

    def testPasswordSettingAll_maybeSamba_WillNotFindAnything(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                )
            ],
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.setPassword(newPasswd=b"new")

        def checkError(fail):
            fail.trap(ldapsyntax.PasswordSetAggregateError)
            value = fail.value
            self.assertEqual(
                str(value),
                "Some of the password plugins failed: "
                "Samba failed with cn=foo,dc=example,dc=com.",
            )
            l = value.errors
            assert len(l) == 1
            assert len(l[0]) == 2
            assert l[0][0] == "Samba"
            assert isinstance(l[0][1], failure.Failure)
            l[0][1].trap(ldapsyntax.DNNotPresentError)
            return "All checks are fine"

        d.addErrback(checkError)

        self.assertEqual("All checks are fine", self.successResultOf(d))
        client.assertSent(
            pureldap.LDAPPasswordModifyRequest(
                userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
            ),
            pureldap.LDAPSearchRequest(
                baseObject="cn=foo,dc=example,dc=com",
                scope=pureldap.LDAP_SCOPE_baseObject,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilterMatchAll,
                attributes=("objectClass",),
            ),
        )

    def testPasswordSetting_abortsOnFirstError(self):
        """LDAPEntry.setPassword() aborts on first error (does not parallelize, as it used to)."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPInsufficientAccessRights.resultCode,
                    matchedDN="",
                    errorMessage="",
                )
            ],
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["foo", "sambaAccount"],
            },
            complete=1,
        )
        d = o.setPassword(newPasswd=b"new")

        def eb(fail):
            fail.trap(ldapsyntax.PasswordSetAggregateError)
            value = fail.value
            self.assertEqual(
                str(value),
                "Some of the password plugins failed: "
                "ExtendedOperation failed with insufficientAccessRights; "
                "Samba failed with Aborted.",
            )
            l = value.errors
            assert len(l) == 2

            assert len(l[0]) == 2
            self.assertEqual(l[0][0], "ExtendedOperation")
            assert isinstance(l[0][1], failure.Failure)
            l[0][1].trap(ldaperrors.LDAPInsufficientAccessRights)

            assert len(l[1]) == 2
            self.assertEqual(l[1][0], "Samba")
            assert isinstance(l[1][1], failure.Failure)
            l[1][1].trap(ldapsyntax.PasswordSetAborted)

            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=foo,dc=example,dc=com", newPasswd=b"new"
                ),
            )

        d.addCallbacks(testutil.mustRaise, eb)
        return d


class LDAPSyntaxFetch(unittest.TestCase):
    def testFetch_WithDirtyJournal(self):
        """Trying to fetch attributes with a dirty journal fails."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        o["x"] = ["foo"]

        self.failUnlessRaises(ldapsyntax.ObjectDirtyError, o.fetch)

    def testFetch_Empty(self):
        """Fetching attributes for a newly-created object works."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = o.fetch()

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                )
            )

            has = o.keys()
            has.sort()
            want = [b"foo", b"bar"]
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o["foo"], [b"a"])
            self.assertEqual(o["bar"], [b"b", b"c"])

        d.addCallback(cb)
        return d

    def testFetch_Prefilled(self):
        """Fetching attributes for a (partially) known object overwrites the old attributes."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        ("foo", ["a"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={"foo": ["x"], "quux": ["baz", "xyzzy"]},
        )
        d = o.fetch()

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                )
            )

            has = o.keys()
            has.sort()
            want = [b"foo", b"bar"]
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o["foo"], [b"a"])
            self.assertEqual(o["bar"], [b"b", b"c"])

        d.addCallback(cb)
        return d

    def testFetch_Partial(self):
        """Fetching only some of the attributes does not overwrite existing values of different attribute types."""
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=foo,dc=example,dc=com",
                    attributes=(
                        (b"foo", [b"a"]),
                        (b"bar", [b"b", b"c"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={b"foo": [b"x"], b"quux": [b"baz", b"xyzzy"]},
        )
        d = o.fetch(b"foo", b"bar", b"thud")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                    attributes=(b"foo", b"bar", b"thud"),
                )
            )

            has = o.keys()
            has.sort()
            want = [b"foo", b"bar", b"quux"]
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o[b"foo"], [b"a"])
            self.assertEqual(o[b"bar"], [b"b", b"c"])
            self.assertEqual(o[b"quux"], [b"baz", b"xyzzy"])

        d.addCallback(cb)
        return d

    def testCommitAndFetch(self):
        """Fetching after a commit works."""

        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0, matchedDN="", errorMessage="")],
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com",
                    [("aValue", ["foo", "bar"])],
                ),
                pureldap.LDAPSearchResultDone(resultCode=0),
            ],
        )
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
            },
        )

        o["aValue"] = ["foo", "bar"]
        d = o.commit()
        d.addCallback(self.assertIdentical, o)

        d.addCallback(lambda _: o.fetch("aValue"))
        d.addCallback(self.assertIdentical, o)

        def cb(dummy):
            client.assertSent(
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Replace("aValue", ["foo", "bar"]),
                    ],
                ).asLDAP(),
                pureldap.LDAPSearchRequest(
                    baseObject="cn=foo,dc=example,dc=com",
                    scope=pureldap.LDAP_SCOPE_baseObject,
                    attributes=["aValue"],
                ),
            )

        d.addCallback(cb)
        return d


class LDAPSyntaxRDNHandling(unittest.TestCase):
    def testRemovingRDNFails(self):
        """Removing RDN fails with CannotRemoveRDNError."""
        o = ldapsyntax.LDAPEntry(
            client=None,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["someObjectClass"],
                "cn": ["foo", "bar", "baz"],
                "a": ["aValue"],
            },
        )
        o["cn"].remove("bar")
        del o["a"]
        self.assertRaisesRegex(
            ldapsyntax.CannotRemoveRDNError,
            re.escape(
                "The attribute to be removed, 'cn'='foo', "
                "is the RDN for the object and cannot be removed."
            ),
            o["cn"].remove,
            "foo",
        )

        def f():
            del o["cn"]

        self.assertRaisesRegex(
            ldapsyntax.CannotRemoveRDNError,
            re.escape(
                "The attribute to be removed, 'cn', "
                "is the RDN for the object and cannot be removed."
            ),
            f,
        )

        def f():
            o["cn"] = ["thud"]

        self.assertRaisesRegex(
            ldapsyntax.CannotRemoveRDNError,
            re.escape(
                "The attribute to be removed, 'cn', "
                "is the RDN for the object and cannot be removed."
            ),
            f,
        )

        # TODO maybe this should be ok, it preserves the RDN.
        # For now, disallow it.
        def f():
            o["cn"] = ["foo"]

        self.assertRaisesRegex(
            ldapsyntax.CannotRemoveRDNError,
            re.escape(
                "The attribute to be removed, 'cn', "
                "is the RDN for the object and cannot be removed."
            ),
            f,
        )


class LDAPSyntaxMove(unittest.TestCase):
    def test_move(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPModifyDNResponse(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["foo"],
                "aValue": ["a"],
            },
        )
        d = o.move("cn=bar,ou=somewhere,dc=example,dc=com")

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPModifyDNRequest(
                    entry="cn=foo,dc=example,dc=com",
                    newrdn="cn=bar",
                    deleteoldrdn=1,
                    newSuperior="ou=somewhere,dc=example,dc=com",
                )
            )

            self.assertEqual(o.dn, "cn=bar,ou=somewhere,dc=example,dc=com")

        d.addCallback(cb)
        return d


class Bind(unittest.TestCase):
    def test_ok(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPBindResponse(resultCode=0, matchedDN=""),
            ]
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = defer.maybeDeferred(o.bind, "s3krit")
        d.addCallback(self.assertIdentical, o)

        def cb(dummy):
            client.assertSent(
                pureldap.LDAPBindRequest(dn="cn=foo,dc=example,dc=com", auth="s3krit")
            )

        d.addCallback(cb)
        return d

    def test_fail(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode,
                    matchedDN="",
                ),
            ]
        )

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = defer.maybeDeferred(o.bind, "s3krit")

        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_err(self):
        client = LDAPClientTestDriver([failure.Failure(error.ConnectionLost())])

        o = ldapsyntax.LDAPEntry(client=client, dn="cn=foo,dc=example,dc=com")
        d = defer.maybeDeferred(o.bind, "whatever")

        def eb(fail):
            fail.trap(error.ConnectionLost)

        d.addCallbacks(testutil.mustRaise, eb)
        return d



from zope.interface import Interface


class ILDAPEntry(Interface):
    """

    Pythonic API for LDAP object access and modification.

    >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
    ...     dn='cn=foo,dc=example,dc=com',
    ...     attributes={'anAttribute': ['itsValue', 'secondValue'],
    ...     'onemore': ['aValue'],
    ...     })
    >>> o
    LDAPEntry(dn='cn=foo,dc=example,dc=com', attributes={'anAttribute': ['itsValue', 'secondValue'], 'onemore': ['aValue']})
    """

    def __getitem__(self, key):
        """

        Get all values of an attribute.

        >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
        ...     dn='cn=foo,dc=example,dc=com',
        ...     attributes={'anAttribute': ['itsValue']})
        >>> o['anAttribute']
        ['itsValue']

        """

    def get(self, key, default=None):
        """

        Get all values of an attribute.

        >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
        ...     dn='cn=foo,dc=example,dc=com',
        ...     attributes={'anAttribute': ['itsValue']})
        >>> o.get('anAttribute')
        ['itsValue']
        >>> o.get('foo')
        >>> o.get('foo', [])
        []

        """

    def has_key(self, key):
        """TODO"""

    def __contains__(self, key):
        """TODO"""

    def keys(self):
        """TODO"""

    def items(self):
        """TODO"""

    def __str__(self):
        """

        Stringify as LDIF.

        >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
        ...     dn='cn=foo,dc=example,dc=com',
        ...     attributes={'anAttribute': ['itsValue', 'secondValue'],
        ...     'onemore': ['aValue'],
        ...     })
        >>> # must use rstrip or doctests won't like it due to the empty line
        >>> # you can just say "print o"
        >>> print str(o).rstrip()
        dn: cn=foo,dc=example,dc=com
        anAttribute: itsValue
        anAttribute: secondValue
        onemore: aValue

        """

    def __eq__(self, other):
        """

        Comparison. Only equality is supported.

        >>> client=ldapclient.LDAPClient()
        >>> a=LDAPEntry(client=client,
        ...             dn='dc=example,dc=com')
        >>> b=LDAPEntry(client=client,
        ...             dn='dc=example,dc=com')
        >>> a==b
        1
        >>> c=LDAPEntry(client=ldapclient.LDAPClient(),
        ...             dn='ou=different,dc=example,dc=com')
        >>> a==c
        0

        Comparison does not consider the client of the object.

        >>> anotherClient=ldapclient.LDAPClient()
        >>> d=LDAPEntry(client=anotherClient,
        ...             dn='dc=example,dc=com')
        >>> a==d
        1

        """

    def __ne__(self, other):
        """

        Inequality comparison. See L{__eq__}.

        """

    def __len__(self):
        """TODO"""

    def __nonzero__(self):
        """Always return True"""

    def bind(self, password):
        """
        Try to authenticate with given secret.

        @return: Deferred ILDAPEntry (that is, self).

        @raise ldaperrors.LDAPInvalidCredentials: password was
        incorrect.
        """


class IEditableLDAPEntry(Interface):
    """Interface definition for editable LDAP entries."""

    def __setitem__(self, key, value):
        """

        Set values of an attribute. Please use lists. Do not modify
        the lists in place, that's not supported _yet_.

        >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
        ...     dn='cn=foo,dc=example,dc=com',
        ...     attributes={'anAttribute': ['itsValue']})
        >>> o['anAttribute']=['foo', 'bar']
        >>> o['anAttribute']
        ['bar', 'foo']

        """

    def __delitem__(self, key):
        """

        Delete all values of an attribute.

        >>> o=LDAPEntry(client=ldapclient.LDAPClient(),
        ...     dn='cn=foo,dc=example,dc=com',
        ...     attributes={
        ...     'anAttribute': ['itsValue', 'secondValue'],
        ...     'another': ['moreValues'],
        ...     })
        >>> del o['anAttribute']
        >>> o
        LDAPEntry(dn='cn=foo,dc=example,dc=com', attributes={'another': ['moreValues']})

        """

    def undo(self):
        """
        Forget all pending changes.
        """

    def commit(self):
        """
        Send all pending changes to the LDAP server.

        @returns: a Deferred that fires True (operation succeeded)
        or False (operation failed).
        """

    def move(self, newDN):
        """

        Move the object to a new DN.

        @param newDN: the new DistinguishedName

        @return: A Deferred that will complete when the move is done.

        """

    def delete(self):
        """

        Delete this object from the LDAP server.

        @return: A Deferred that will complete when the delete is done.

        """

    def setPassword(self, newPasswd):
        """

        Set all applicable passwords for this object.

        @param newPasswd: A string containing the new password.

        @return: A Deferred that will complete when the operation is
        done.

        """


class IConnectedLDAPEntry(Interface):
    """
    Interface definition for LDAP entries that are part of a bigger
    whole.
    """

    def namingContext(self):
        """
        Return an LDAPEntry for the naming context that contains this object.
        """

    def fetch(self, *attributes):
        """
        Fetch the attributes of this object from the server.

        @param attributes: Attributes to fetch. If none, fetch all
        attributes. Fetched attributes are overwritten, and if
        fetching all attributes, attributes that are not on the server
        are removed.

        @return: A Deferred that will complete when the operation is
        done.
        """

    def search(
        self,
        filterText=None,
        filterObject=None,
        attributes=(),
        scope=None,
        derefAliases=None,
        sizeLimit=0,
        timeLimit=0,
        typesOnly=0,
        callback=None,
    ):
        """

        Perform an LDAP search with this object as the base.

        @param filterText: LDAP search filter as a string.

        @param filterObject: LDAP search filter as LDAPFilter.
        Note if both filterText and filterObject are given, they
        are combined with AND. If neither is given, the search is
        made with a filter that matches everything.

        @param attributes: List of attributes to retrieve for the
        result objects. An empty list and means all.

        @param scope: Whether to recurse into subtrees.

        @param derefAliases: Whether to deref LDAP aliases. TODO write
        better documentation.

        @param sizeLimit: At most how many entries to return. 0 means
        unlimited.

        @param timeLimit: At most how long to use for processing the
        search request. 0 means unlimited.

        @param typesOnly: Whether to return attribute types only, or
        also values.

        @param callback: Callback function to call for each resulting
        LDAPEntry. None means gather the results into a list and give
        that to the Deferred returned from here.

        @return: A Deferred that will complete when the search is
        done. The Deferred gives None if callback was given and a list
        of the search results if callback is not given or is None.

        """

    def children(self, callback=None):
        """

        List the direct children of this entry. Try to avoid using
        .search(), as this will be used later to implement .search()
        on LDAP backends.

        @param callback: Callback function to call for each resulting
        LDAPEntry. None means gather the results into a list and give
        that to the Deferred returned from here.

        @return: A Deferred that will complete when the list is
        over. The Deferred gives None if callback was given and a list
        of the children if callback is not given or is None.

        """

    def subtree(self, callback=None):
        """

        List the subtree rooted at this entry, including this
        entry. Try to avoid using .search(), as this will be used
        later to implement .search() on LDAP backends.

        @param callback: Callback function to call for each resulting
        LDAPEntry. None means gather the results into a list and give
        that to the Deferred returned from here.

        @return: A Deferred that will complete when the list is
        over. The Deferred gives None if callback was given and a list
        of the children if callback is not given or is None.

        """

    def lookup(self, dn):
        """
        Lookup the referred to by dn.

        @return: A Deferred returning an ILDAPEntry, or failing with e.g.
        LDAPNoSuchObject.
        """

    def match(self, filter):
        """

        Does entry match filter.

        @param filter: An LDAPFilter (e.g. LDAPFilter_present,
        LDAPFilter_equalityMatch etc. TODO provide an interface or
        superclass for filters.)

        @return: Boolean.

        """


class ILDAPConfig(Interface):
    """Generic LDAP configuration retrieval."""

    def getBaseDN(self):
        """
        Get the LDAP base DN, as a DistinguishedName.

        Raises ldaptor.config.MissingBaseDNError
        if configuration does not specify a base DN.
        """

    def getServiceLocationOverrides(self):
        """
        Get the LDAP service location overrides, as a mapping of
        DistinguishedName to (host, port) tuples.
        """

    def copy(self, baseDN=None, serviceLocationOverrides=None):
        """
        Make a copy of this configuration, overriding certain aspects
        of it.
        """

    def getIdentityBaseDN(self):
        """TODO"""

    def getIdentitySearch(self, name):
        """TODO"""



"""
Test cases for the ldaptor.config module.
"""

import os

from twisted.trial import unittest
from ldaptor import config


def writeFile(path, content):
    f = open(path, "wb")
    f.write(content)
    f.close()


def reloadFromContent(testCase, content):
    """
    Reload the global configuration file with raw `content`.
    """
    base_path = testCase.mktemp()
    os.mkdir(base_path)
    config_path = os.path.join(base_path, "test.cfg")
    writeFile(config_path, content)

    # Reload with empty content to reduce the side effects.
    testCase.addCleanup(reloadFromContent, testCase, b"")

    return config.loadConfig(
        configFiles=[config_path],
        reload=True,
    )


class TestLoadConfig(unittest.TestCase):
    """
    Tests for loadConfig.
    """

    def testMultileConfigurationFile(self):
        """
        It can read configuration from multiple files, merging the
        loaded values.
        """
        self.dir = self.mktemp()
        os.mkdir(self.dir)
        self.f1 = os.path.join(self.dir, "one.cfg")
        writeFile(
            self.f1,
            b"""\
[fooSection]
fooVar = val

[barSection]
barVar = anotherVal
""",
        )
        self.f2 = os.path.join(self.dir, "two.cfg")
        writeFile(
            self.f2,
            b"""\
[fooSection]
fooVar = val2
""",
        )
        self.cfg = config.loadConfig(configFiles=[self.f1, self.f2], reload=True)

        val = self.cfg.get("fooSection", "fooVar")
        self.assertEqual(val, "val2")

        val = self.cfg.get("barSection", "barVar")
        self.assertEqual(val, "anotherVal")


class TestLDAPConfig(unittest.TestCase):
    """
    Unit tests for LDAPConfig.
    """

    def testGetBaseDNOK(self):
        """
        It will return the base DN found in the configuration in the [ldap]
        section as `base` option.
        """
        reloadFromContent(self, b"[ldap]\nbase=dc=test,dc=net\n")
        sut = config.LDAPConfig()

        result = sut.getBaseDN()

        self.assertEqual("dc=test,dc=net", result)

    def testGetBaseDNNoSection(self):
        """
        It raise an exception when the the configuration has no [ldap]
        section.
        """
        reloadFromContent(self, b"[other]\nbase=dc=test,dc=net\n")
        sut = config.LDAPConfig()

        self.assertRaises(
            config.MissingBaseDNError,
            sut.getBaseDN,
        )

    def testGetBaseDNNoOption(self):
        """
        It raise an exception when the the configuration has [ldap]
        section but no `base` option.
        """
        reloadFromContent(self, b"[ldap]\nbaseless=dc=test,dc=net\n")
        sut = config.LDAPConfig()

        self.assertRaises(
            config.MissingBaseDNError,
            sut.getBaseDN,
        )

    def testGetIdentityBaseDNOK(self):
        """
        It will return the value found in the configuration in the
        [authentication] section as `identity-base` option.
        """
        reloadFromContent(
            self, b"[authentication]\n" b"identity-base=ou=users,dc=test,dc=net\n"
        )
        sut = config.LDAPConfig()

        result = sut.getIdentityBaseDN()

        self.assertEqual("ou=users,dc=test,dc=net", result)

    def testGetIdentityBaseSectionSection(self):
        """
        When the configuration does not contains the
        `[authentication]` section it will return the configured Base DN.
        """
        reloadFromContent(self, b"[ldap]\n" b"basE=dc=test,dc=net\n")
        sut = config.LDAPConfig()

        result = sut.getIdentityBaseDN()

        self.assertEqual("dc=test,dc=net", result)

    def testGetIdentityBaseNoOption(self):
        """
        When the configuration does not contains the `identity-base` option
        inside the `[authentication]` section it will return the configured
        Base DN.
        """
        reloadFromContent(
            self,
            b"[ldap]\n"
            b"BASE=dc=test,dc=net\n"
            b"[authentication]\n"
            b"no-identity-base=dont care\n",
        )
        sut = config.LDAPConfig()

        result = sut.getIdentityBaseDN()

        self.assertEqual("dc=test,dc=net", result)

    def testGetIdentitySearchOK(self):
        """
        It will use the value from to configuration for its return value.
        """
        reloadFromContent(
            self,
            b"""[authentication]
identity-search = (something=%(name)s)
""",
        )
        sut = config.LDAPConfig()

        result = sut.getIdentitySearch("foo")

        self.assertEqual("(something=foo)", result)

    def testGetIdentitySearchNoSection(self):
        """
        When the configuration file does not contains the `authentication`
        section it will use a default expression.
        """
        sut = config.LDAPConfig()

        result = sut.getIdentitySearch("foo")

        self.assertEqual("(|(cn=foo)(uid=foo))", result)

    def testGetIdentitySearchNoOption(self):
        """
        When the configuration file contains the `authentication`
        section but without the identity search option,
        it will use a default expression.
        """
        reloadFromContent(self, b"[authentication]\nother_key=value")
        sut = config.LDAPConfig()

        result = sut.getIdentitySearch("foo")

        self.assertEqual("(|(cn=foo)(uid=foo))", result)

    def testgetIdentitySearchFromInitArguments(self):
        """
        When data is provided at LDAPConfig initialization it is used
        as the backend data.
        """
        sut = config.LDAPConfig(identitySearch="(&(bar=thud)(quux=%(name)s))")

        result = sut.getIdentitySearch("foo")

        self.assertEqual("(&(bar=thud)(quux=foo))", result)

    def testCopy(self):
        """
        It returns a copy of the configuration.
        """
        sut = config.LDAPConfig()

        copied = sut.copy(identitySearch="(&(bar=baz)(quux=%(name)s))")

        self.assertIsInstance(copied, config.LDAPConfig)

        result = copied.getIdentitySearch("foo")

        self.assertEqual("(&(bar=baz)(quux=foo))", result)



from twisted.internet import reactor
from twisted.web import server, resource

from ldaptor.protocols.ldap import (
    ldapclient,
    ldapsyntax,
    ldapconnector,
    distinguishedname,
)
from ldaptor import ldapfilter


class LDAPConfig:
    def __init__(self, baseDN, serviceLocationOverrides=None):
        self.baseDN = distinguishedname.DistinguishedName(baseDN)
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k, v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn] = v

    def getBaseDN(self):
        return self.baseDN

    def getServiceLocationOverrides(self):
        return self.serviceLocationOverrides


class AddressBookResource(resource.Resource):
    def __init__(self, config):
        resource.Resource.__init__(self)
        self.config = config
        self.putChild("", self)

    def _search(self, searchFilter):
        c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d = c.connectAnonymously(
            self.config.getBaseDN(), self.config.getServiceLocationOverrides()
        )

        def _doSearch(proto, searchFilter):
            searchFilter = ldapfilter.parseFilter(searchFilter)
            baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=self.config.getBaseDN())
            d = baseEntry.search(filterObject=searchFilter)
            return d

        d.addCallback(_doSearch, searchFilter)
        return d

    def _show(self, results, write):
        for item in results:
            write("<pre>")
            write(str(item))
            write("</pre>")

    def render(self, request):
        searchFilter = request.args.get("search", [None])[0]
        if not searchFilter:
            return "Please give a search filter."
        d = self._search(searchFilter)
        d.addCallback(self._show, request.write)
        d.addCallback(
            lambda _: request.write("<p>Search filter was %s</p>" % searchFilter)
        )
        d.addErrback(lambda e: request.write(str(e)))
        d.addBoth(lambda _: request.finish())
        return server.NOT_DONE_YET


def getSite(config):
    return server.Site(AddressBookResource(config))



"""
Test cases for ldaptor.entry
"""

from twisted.trial import unittest

from ldaptor import delta, entry
from ldaptor.protocols.ldap.ldaperrors import LDAPInvalidCredentials


class TestBaseLDAPEntry(unittest.TestCase):
    """
    Tests for ldaptor.entry.BaseLDAPEntry.
    """

    def testEqualitySameType(self):
        """
        It is equal if it has the same DN (case insensitive)
        and same attributes and values.
        """
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="Dc=Foo",
            attributes={
                "foo": ["bar"],
            },
        )

        self.assertEqual(a, b)

    def testEqualityDifferentType(self):
        """
        It is not equal with objects of different types.
        """
        a = entry.BaseLDAPEntry(dn="dc=foo", attributes={})
        self.assertFalse(a == object())

    def testInequalityDifferentDN(self):
        """
        Entries are not equal if their DNs are not equal
        """
        a = entry.BaseLDAPEntry(dn="dn=foo", attributes={"foo": ["bar"]})
        b = entry.BaseLDAPEntry(dn="dn=bar", attributes={"foo": ["bar"]})
        self.assertNotEqual(a, b)

    def testBindPlainText(self):
        """
        It will bind when the password for the entry is stored in plain text,
        and will return a deferred which has itself as callback.
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "userPassword": [b"some-plain-text"],
            },
        )

        deferred = sut.bind(b"some-plain-text")
        result = self.successResultOf(deferred)

        self.assertIs(sut, result)

    def testBindSeededSHA(self):
        """
        It can bind with password stored in seeded SHA.
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "userPassword": [b"{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="],
            },
        )

        deferred = sut.bind(b"secret")
        result = self.successResultOf(deferred)

        self.assertIs(sut, result)

    def testBindPlainTextError(self):
        """
        Return a LDAPInvalidCredentials failure when password don't match.
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "userPassword": [b"some-plain-text"],
            },
        )

        deferred = sut.bind(b"other-password")
        failure = self.failureResultOf(deferred)

        self.assertTrue(failure.check(LDAPInvalidCredentials))

    def testBindSHAError(self):
        """
        Return a LDAPInvalidCredentials failure when encoded password don't
        match.
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "userPassword": [b"{SSHA}anythinghere"],
            },
        )

        deferred = sut.bind(b"other-password")
        failure = self.failureResultOf(deferred)

        self.assertTrue(failure.check(LDAPInvalidCredentials))

    def testGetLDIF(self):
        """
        Getting human readable representation of an entry
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "bar": ["foo"],
            },
        )
        self.assertEqual(sut.getLDIF(), "dn: dc=foo\nbar: foo\nfoo: bar\n\n")

    def testNonzero(self):
        """Entry is always non-zero"""
        sut = entry.BaseLDAPEntry(dn="")
        self.assertTrue(bool(sut))

    def testRepr(self):
        """
        Getting string representation of an entry
        """
        sut = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "bar": ["foo"],
            },
        )
        self.assertEqual(
            repr(sut), "BaseLDAPEntry('dc=foo', {'bar': ['foo'], 'foo': ['bar']})"
        )


class TestDiffEntry(unittest.TestCase):
    """
    Tests for ldaptor.entry.BaseLDAPEntry.diff()
    """

    def testEqual(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        result = a.diff(b)
        self.assertEqual(result, None)

    def testAdd_New_OneType_OneValue(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("baz", ["quux"]),
                ],
            ),
        )

    def testAdd_New_OneType_ManyValues(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux", "thud", "foo"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("baz", ["quux", "thud", "foo"]),
                ],
            ),
        )

    def testAdd_New_ManyTypes(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux"],
                "bang": ["thud"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("bang", ["thud"]),
                    delta.Add("baz", ["quux"]),
                ],
            ),
        )

    def testAdd_Existing_OneType_OneValue(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar", "quux"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("foo", ["quux"]),
                ],
            ),
        )

    def testAdd_Existing_OneType_ManyValues(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar", "quux", "thud", "foo"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("foo", ["quux", "thud", "foo"]),
                ],
            ),
        )

    def testAdd_NewAndExisting_ManyTypes(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar", "thud", "bang"],
                "baz": ["quux", "bar", "stump"],
                "bang": ["thud", "barble"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Add("bang", ["thud", "barble"]),
                    delta.Add("baz", ["bar", "stump"]),
                    delta.Add("foo", ["thud", "bang"]),
                ],
            ),
        )

    def testDelete_All_OneType(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux", "thud"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Delete("baz", ["quux", "thud"]),
                ],
            ),
        )

    def testDelete_Some_OneType(self):
        a = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["quux", "thud"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="dc=foo",
            attributes={
                "foo": ["bar"],
                "baz": ["thud"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "dc=foo",
                [
                    delta.Delete("baz", ["quux"]),
                ],
            ),
        )

    def testComplex(self):
        a = entry.BaseLDAPEntry(
            dn="cn=Paula Jensen,ou=Product Development,dc=airius,dc=com",
            attributes={
                "description": ["Something"],
                "telephonenumber": ["+123 456"],
                "facsimiletelephonenumber": ["+1 408 555 9876"],
            },
        )
        b = entry.BaseLDAPEntry(
            dn="cn=Paula Jensen,ou=Product Development,dc=airius,dc=com",
            attributes={
                "postalAddress": ["123 Anystreet $ Sunnyvale, CA $ 94086"],
                "telephonenumber": ["+1 408 555 1234", "+1 408 555 5678"],
            },
        )
        result = a.diff(b)
        self.assertEqual(
            result,
            delta.ModifyOp(
                "cn=Paula Jensen,ou=Product Development,dc=airius,dc=com",
                [
                    delta.Add(
                        "postalAddress", ["123 Anystreet $ Sunnyvale, CA $ 94086"]
                    ),
                    delta.Delete("description", ["Something"]),
                    delta.Delete("facsimiletelephonenumber", ["+1 408 555 9876"]),
                    delta.Add(
                        "telephonenumber", ["+1 408 555 1234", "+1 408 555 5678"]
                    ),
                    delta.Delete("telephonenumber", ["+123 456"]),
                ],
            ),
        )



from passlib.hash import nthash as passlib_nthash, lmhash as passlib_lmhash

from ldaptor import config


def nthash(password=b""):
    """Generates nt md4 password hash for a given password."""
    return passlib_nthash.hash(password[:128]).encode("ascii").upper()


def lmhash_locked(password=b""):
    """
    Generates a lanman password hash that matches no password.

    Note that the author thinks LanMan hashes should be banished from
    the face of the earth.
    """
    return 32 * b"X"


def lmhash(password=b""):
    """
    Generates lanman password hash for a given password.

    Note that the author thinks LanMan hashes should be banished from
    the face of the earth.
    """

    if not config.useLMhash():
        return lmhash_locked()

    return passlib_lmhash.hash(password).encode("ascii").upper()
